***************************************
*
*     PROJECT: Faces detection (Xbase++ integration of fdlib.dll from Wolf Kienzle)
*        FILE: Faces.prg
*
*              ( With code from Jack Duijf & Pablo Botella: _WEBCAMDIALOG.PRG, WEBCAMDIALOG.PRG, WindowsDll.prg, XbpWebcam.prg )
*
*  Programmer: J.A. Diego Kerejeta
*     Created: 17-03-2015
*  
*
*     Version: 0.2
*
****************************************


#include 'xbp.ch'
#include 'common.ch'
#include 'AppEvent.ch'
#include 'dll.ch'
#include 'gra.ch'

#include 'c:\ot4xb\ot4xb.ch'                 // ****** IMPORTANT: ADAPT THE PATH *******
#pragma library ( "C:\OT4XB\OT4XB.LIB" )     // ****** IMPORTANT: ADAPT THE PATH *******

#define DEVICE_ID       1
#define DEVICE_NAME     2
#define DEVICE_VERSION  3

* #define SHOW_GRAY_STEPS                    // Activate for test purpose

#define FACE_TOP_MARGIN              1
#define FACE_LEFT_RIGHT_MARGIN       2
#define FACE_BOTTOM_MARGIN           3

#define FACE_FIND_MSG_0              '  No face detected'
#define FACE_FIND_MSG_1              '  Face detected'
#define FACE_FIND_MSG_2              '  More than one face detected'


PROCEDURE AppSys() ; SET CHARSET TO ANSI ; RETURN
******************
PROCEDURE DbeSys() ; RETURN
******************

**************
Procedure Main
**************
LOCAL aDevices:= LoadDeviceList()
LOCAL oDlg, nEvent, mp1, mp2, oXbp

IF Len( aDevices ) > 0
   oDlg:= WebcamDialog():new()
      oDlg:setDevice( aDevices[ 1, DEVICE_ID ], aDevices[ 1, DEVICE_NAME ], aDevices[ 1, DEVICE_VERSION ] )
      oDlg:title:= nil
      oDlg:create()

      mp1:= oDlg:oImage:currentResolution()

      oDlg:setTitle( 'Faces detection (current resolution: ' + Var2Char( mp1[ 1 ] ) + 'x' + Var2char( mp1[ 2 ] ) + ')' )

      oDlg:oBtnSave:activate   := {|| Save( oDlg ) }
      oDlg:oBtnCapture:activate:= {|| Capture( oDlg ) }

   SetAppWindow( oDlg )
   SetAppFocus( oDlg:oBtnCapture )

   DO WHILE nEvent <> xbeP_Close
      nEvent      := AppEvent( @mp1, @mp2, @oXbp )
      oXbp:handleEvent( nEvent, mp1, mp2 )
   ENDDO
ELSE
   MsgBox( "No capture devices available" )
ENDIF
RETURN

***********************
Function LoadDeviceList
***********************
LOCAL cName:= ChrR(0,260)
LOCAL cVer := ChrR(0,260)
LOCAL aDev := {}
LOCAL n 

FOR n := 0 to 9
   IF capGetDriverDescription( n, @cName, Len( cName ), @cVer, Len( cVer ) )
      aadd( aDev, { n, PeekStr( cName,, -1 ), PeekStr( cVer,, -1 ) } )
   ENDIF
NEXT
RETURN aDev

***************************
PROCEDURE Save( oDlgOwner )
***************************
LOCAL cFile:= StrTran( AppName( .T. ), AppName(), '' ) + 'Face.jpg'

Ferase( cFile )
oDlgOwner:oFace:caption:saveFile( cFile, XBPBMP_FORMAT_JPG )
RETURN

******************************
PROCEDURE Capture( oDlgOwner )
******************************
LOCAL oBmp
LOCAL nThreshold       := oDlgOwner:oSpinThreshold:getData()
LOCAL nRejectedMinSize := oDlgOwner:oSpinRejectSize:getData()
LOCAL aFaceFinalSize   := { oDlgOwner:oSpinSizeX:getData(), oDlgOwner:oSpinSizeY:getData() }
LOCAL lFaceEqualization:= oDlgOwner:oChkEqualize:getData()
LOCAL aMarginFactors   := { oDlgOwner:oSpinFactor1:getData() / 10, oDlgOwner:oSpinFactor2:getData() / 10, ;
                            oDlgOwner:oSpinFactor3:getData() / 10 }

oDlgOwner:oStatus:setCaption( '' )

oBmp:= oDlgOwner:oImage:capture()

oDlgOwner:oBitmap:setCaption( oBmp )

IF FindFace( oDlgOwner, oBmp, nRejectedMinSize, nThreshold, aFaceFinalSize, aMarginFactors, lFaceEqualization )
   oDlgOwner:oBtnSave:enable()
ENDIF
RETURN 

*****************************************************************************************************************
FUNCTION FindFace( oDlgOwner, oBmp, nFaceMinSize, nThreshold, aFaceFinalSize, aMarginFactors, lFaceEqualization )
*****************************************************************************************************************
LOCAL nFaces, nXPos, nYPos, nSize, nCont
LOCAL cGrayData   := ExtractGrayData( oBmp, oDlgOwner:oBitmap )
LOCAL nXSize      := oBmp:xsize
LOCAL nYSize      := oBmp:ysize
LOCAL nLen        := nXSize * nYSize
LOCAL cInvGrayData:= ''
LOCAL aFaces      := {}

#ifdef SHOW_GRAY_STEPS ; ShowBmpGrayStep( cGrayData, oBmp, oDlgOwner:oBitmap ) ; #endif

FOR nCont:= 1 TO nLen STEP nXSize
   cInvGrayData:= SubStr( cGrayData, nCont, nXSize ) + cInvGrayData
NEXT

fdlib_detectfaces( cInvGrayData, nXSize, nYSize, nThreshold )
	
nFaces:= fdlib_getndetections()

FOR nCont:= 1 TO nFaces

   nXPos:= nYPos:= nSize:= 0

   fdlib_getdetection( nCont - 1, @nXPos, @nYPos, @nSize )

   IF nXPos > 0 .AND. ( nFaceMinSize = nil .OR. nFaceMinSize < nSize )
      AAdd( aFaces, { nXPos, nYPos, nSize } )
   ENDIF
NEXT
nFaces:= Len( aFaces )

IF nFaces = 0
   oDlgOwner:oStatus:setCaption( FACE_FIND_MSG_0 )
ELSEIF nFaces > 1 
   oDlgOwner:oStatus:setCaption( FACE_FIND_MSG_2 )
ELSE
   oDlgOwner:oStatus:setCaption( FACE_FIND_MSG_1 )
   ShowFace( aFaces[ 1 ], oBmp, oDlgOwner:oBitmap, oDlgOwner:oFace, aFaceFinalSize, aMarginFactors, lFaceEqualization )
   RETURN .T.
ENDIF
RETURN .F.

**************************************************************************************************************************
STATIC PROCEDURE ShowFace( aFace, oBmp, oXbpShowCapture, oXbpShowFace, aFaceFinalSize, aMarginFactors, lFaceEqualization )
**************************************************************************************************************************
LOCAL oPS, aParentSize, oFaceBmp
LOCAL nRadius    := Int( aFace[ 3 ] / 2 )
LOCAL nXPos      := aFace[ 1 ]
LOCAL nYPos      := oBmp:ysize - aFace[ 2 ]
LOCAL nX         := Int( nRadius * aMarginFactors[ FACE_LEFT_RIGHT_MARGIN ] )
LOCAL nHeadLeft  := nXPos - nX
LOCAL nHeadRight := nXPos + nX
LOCAL nHeadTop   := nYPos + Int( nRadius * aMarginFactors[ FACE_TOP_MARGIN    ] )
LOCAL nHeadBottom:= nYPos - Int( nRadius * aMarginFactors[ FACE_BOTTOM_MARGIN ] )

AdjustRect( oBmp, @nHeadLeft, @nHeadBottom, @nHeadRight, @nHeadTop, aFaceFinalSize )

oPS:= oXbpShowCapture:lockPS()
   oPS:setcolor( GRA_CLR_RED    ) ; GraBox( oPS, { nXPos - nRadius, nYPos - nRadius }, { nXPos + nRadius, nYPos + nRadius } )
   oPS:setcolor( GRA_CLR_YELLOW ) ; GraBox( oPS, { nHeadLeft, nHeadBottom }, { nHeadRight - 1, nHeadTop - 1 } )
oXbpShowCapture:unlockPS()

oFaceBmp:= BmpCropResize( oBmp, { nHeadLeft, nHeadBottom, nHeadRight, nHeadTop }, aFaceFinalSize )

IF lFaceEqualization ; HistoEqualColor( oFaceBmp ) ; ENDIF

aParentSize:= oXbpShowFace:setParent():currentSize()
oXbpShowFace:hide()
oXbpShowFace:setPosAndSize( { Int( ( aParentSize[ 1 ] - oFaceBmp:xsize ) / 2 ), ;
                              Int( ( aParentSize[ 2 ] - oFaceBmp:ysize ) / 2 ) }, { oFaceBmp:xsize, oFaceBmp:ysize } )
oXbpShowFace:setCaption( oFaceBmp )
oXbpShowFace:show()
RETURN

*************************************************************************************************
STATIC PROCEDURE AdjustRect( oBmp, nHeadLeft, nHeadBottom, nHeadRight, nHeadTop, aFaceFinalSize )
*************************************************************************************************
LOCAL nXSize, nAdjust, _nAdjust
LOCAL nCont:= 0

IF nHeadTop    > oBmp:ysize ; nHeadTop   := oBmp:ysize ; ENDIF
IF nHeadBottom < 0          ; nHeadBottom:= 0          ; ENDIF
IF nHeadLeft   < 0          ; nHeadLeft  := 0          ; ENDIF
IF nHeadRight  > oBmp:xsize ; nHeadRight := oBmp:xsize ; ENDIF

FOR nCont:= 1 TO 400
   nXSize:= nHeadRight - nHeadLeft

   IF nXSize >= aFaceFinalSize[ 1 ]
      IF ( nAdjust:= ( nHeadTop - nHeadBottom ) - Round( nXSize * aFaceFinalSize[ 2 ] / aFaceFinalSize[ 1 ], 0 ) ) = 0
         RETURN 
      ELSEIF nAdjust > 0
         DO WHILE .T.
            nAdjust-- ; nHeadBottom++
            IF nAdjust > 0 ; nAdjust-- ; nHeadTop-- ;  ENDIF
            IF nAdjust = 0 ; RETURN    ; ENDIF
         ENDDO
      ELSE
         nAdjust:= Abs( nAdjust )

         DO WHILE .T.
            _nAdjust:= nAdjust
            IF nAdjust > 0 .AND. nHeadBottom > 0          ; nAdjust-- ; nHeadBottom-- ; ENDIF
            IF nAdjust > 0 .AND. nHeadTop    < oBmp:ysize ; nAdjust-- ; nHeadTop++    ; ENDIF
            IF nAdjust = 0        ; RETURN ; ENDIF
            IF nAdjust = _nAdjust ; EXIT   ; ENDIF
         ENDDO
         nHeadLeft++
         nHeadRight--
      ENDIF
   ELSE
      nHeadLeft := Max( nHeadLeft  - 1, 0 )
      nHeadRight:= Min( nHeadRight + 1, oBmp:xSize )
   ENDIF
NEXT
RETURN 

********************************************************
STATIC FUNCTION ExtractGrayData( oBmp, oXbpShowCapture )
********************************************************
LOCAL nRowPadding, nLastPixel, i, nCont, nGray, aColorTable
LOCAL nBits        := oBmp:bits
LOCAL cBuffer      := oBmp:setBuffer()
LOCAL nBufferOffset:= oBmp:bufferOffset
LOCAL nXSize       := oBmp:xSize
LOCAL nYSize       := oBmp:ySize
LOCAL aHistogram   := AFill( Array( 256 ), 0 )
LOCAL aGrayData    := Array( nXSize * nYSize )
LOCAL nGrayIndex   := 0

#ifdef SHOW_GRAY_STEPS ; LOCAL cGrayData:= '' ; #endif

IF nBits >= 24
   nBits      /= 8
   nRowPadding:= nXSize * nBits % 4
   nLastPixel := nBufferOffset + nXSize * nYSize * nBits + nRowPadding * ( nYSize - 1 )
   nCont      := 0

   FOR i:= nBufferOffset + 1 TO nLastPixel STEP nBits
      nGray                    := Int( 0.0722 * Asc( cBuffer[ i ] ) + 0.7152 * Asc( cBuffer[ i + 1 ] ) + 0.2126 * Asc( cBuffer[ i + 2 ] ) )
      aGrayData[ ++nGrayIndex ]:= nGray
      aHistogram[ nGray + 1 ]++

#ifdef SHOW_GRAY_STEPS ; cGrayData += Chr( nGray ) ; #endif

      IF nRowPadding > 0 .AND. ++nCont = nXSize
         nCont:= 0
         i    += nRowPadding
      ENDIF
   NEXT
ELSEIF nBits >= 1 .AND. nBits < 24
   aColorTable:= oBmp:getColorTable()
   nLastPixel := Len( aColorTable )

   FOR i:= 1 TO nLastPixel
      nGray                    := Int( 0.0722 * aColorTable[ i, 1 ] + 0.7152 * aColorTable[ i, 2 ] + 0.2126 * aColorTable[ i, 3 ] )
      aGrayData[ ++nGrayIndex ]:= nGray
      aHistogram[ nGray + 1 ]++
   NEXT
ENDIF

#ifdef SHOW_GRAY_STEPS ; Sleep( 100 ) ; ShowBmpGrayStep( cGrayData, oBmp, oXbpShowCapture ) ; Sleep( 100 ) ; #endif

RETURN HistoEqualGray( aGrayData, aHistogram )

*******************************************************
STATIC FUNCTION HistoEqualGray( aGrayData, aHistogram )
*******************************************************
LOCAL aCDF:= AFill( Array( 256 ), 0 )
LOCAL aEQU:= AClone( aCDF )
LOCAL nCont, nMin, nMax, nSum, nCDFMin, cGrayData

nCont:= 0
DO WHILE aHistogram[ ++nCont ] = 0 ; ENDDO 
nMin:= nCont 

nCont:= 257
DO WHILE aHistogram[ --nCont ] = 0 ; ENDDO
nMax:= nCont

nCDFMin:= nSum:= aHistogram[ nMin ]

FOR nCont:= nMin + 1 TO nMax
   nSum         += aHistogram[ nCont ]
   aCDF[ nCont ]:= nSum
NEXT
nSum -= nCDFMin

FOR nCont:= nMin TO nMax ; aEQU[ nCont ]:= Chr( Round( 255 * ( aCDF[ nCont ] - nCDFMin ) / nSum, 0 ) ) ; NEXT

nSum     += nCDFMin
cGrayData:= Space( nSum )

FOR nCont:= 1 TO nSum ; cGrayData[ nCont ]:= aEQU[ aGrayData[ nCont ] + 1 ] ; NEXT

RETURN cGrayData

***************************************
STATIC FUNCTION HistoEqualColor( oBmp )
***************************************
LOCAL nRowPadding, cRowPadding, cPixelPadding, nLastPixel, i, nCont
LOCAL aHSV, aPT, nMin, nMax, nSum, nCDFMin
LOCAL nBits        := oBmp:bits
LOCAL cBuffer      := oBmp:setBuffer()
LOCAL nBufferOffset:= oBmp:bufferOffset
LOCAL nXSize       := oBmp:xSize
LOCAL nYSize       := oBmp:ySize
LOCAL aHistogram   := AFill( Array( 256 ), 0 )
LOCAL aCDF         := AClone( aHistogram )
LOCAL aEQU         := AClone( aHistogram )
LOCAL aData        := {}

IF nBits = 24 .OR. nBits = 32
   nBits        /= 8
   nRowPadding  := nXSize * nBits % 4
   cRowPadding  := Replicate( Chr( 0 ), nRowPadding )
   cPixelPadding:= Replicate( Chr( 0 ), nBits - 3  )
   nLastPixel   := nBufferOffset + nXSize * nYSize * nBits + nRowPadding * ( nYSize - 1 )
   nCont        := 0

   FOR i:= nBufferOffset + 1 TO nLastPixel STEP nBits
      aHSV:= Rgb2Hsv( Asc( cBuffer[ i ] ), Asc( cBuffer[ i + 1 ] ), Asc( cBuffer[ i + 2 ] ) )
      AAdd( aData, aHSV )
      aHistogram[ 1 + Round( 255 * aHSV[ 3 ], 0 ) ]++

      IF nRowPadding > 0 .AND. ++nCont = nXSize
         nCont:= 0
         i    += nRowPadding
      ENDIF
   NEXT
ELSE
   RETURN nil
ENDIF

nCont:= 0
DO WHILE aHistogram[ ++nCont ] = 0 ; ENDDO 
nMin:= nCont 

nCont:= 257
DO WHILE aHistogram[ --nCont ] = 0 ; ENDDO
nMax:= nCont

nCDFMin:= nSum:= aHistogram[ nMin ]

FOR nCont:= nMin + 1 TO nMax
   nSum         += aHistogram[ nCont ]
   aCDF[ nCont ]:= nSum
NEXT
nSum -= nCDFMin

FOR nCont:= nMin TO nMax ; aEQU[ nCont ]:= ( aCDF[ nCont ] - nCDFMin ) / nSum ; NEXT

cBuffer   := Left( cBuffer, nBufferOffset )
nCont     := 0
nLastPixel:= Len( aData )

FOR i:= 1 TO nLastPixel
   aPT:= aData[ i ]
   aPT:= Hsv2Rgb( aPT[ 1 ], aPT[ 2 ], aEQU[ 1 + Round( 255 * aPT[ 3 ], 0 ) ] )
   cBuffer += Chr( aPT[ 1 ] ) + Chr( aPT[ 2 ] ) + Chr( aPT[ 3 ] ) + cPixelPadding

   IF nRowPadding > 0 .AND. ++nCont = nXSize
      nCont  := 0
      cBuffer+= cRowPadding
   ENDIF
NEXT
oBmp:setBuffer( cBuffer )
RETURN nil

*************************************
STATIC FUNCTION Rgb2Hsv( nR, nG, nB )
*************************************
LOCAL nMinRGB, nMaxRGB, nD, nH, nMm

nR /= 255 ; nG /= 255 ; nB /= 255

nMinRGB:= Min( nR, Min( nG, nB ) )
nMaxRGB:= Max( nR, Max( nG, nB ) )

IF nMinRGB == nMaxRGB
  RETURN { 0, 0, nMinRGB }
ELSEIF nR = nMinRGB ; nD:= nG - nB ; nH:= 3
ELSEIF nB = nMinRGB ; nD:= nR - nG ; nH:= 1
ELSE                ; nD:= nB - nR ; nH:= 5
ENDIF
nMm:= nMaxRGB - nMinRGB
RETURN { 60 * ( nH - nD / nMm ), nMm / nMaxRGB, nMaxRGB }

*************************************
STATIC FUNCTION Hsv2Rgb( nH, nS, nV )
*************************************
LOCAL nR, nG, nB, nI, nF, nP, nQ, nT

if nS == 0 
   nR:= nG:= nB:= nV
ELSE
   nH /= 60
   nI := Int( nH )
   nF := nH - nI
   nP := nV * ( 1 - nS )
   nQ := nV * ( 1 - nS * nF )
   nT := nV * ( 1 - nS * ( 1 - nF ) )
   IF     nI = 0 ; nR := nV ; nG := nT ; nB := nP
   ELSEIF nI = 1 ; nR := nQ ; nG := nV ; nB := nP
   ELSEIF nI = 2 ; nR := nP ; nG := nV ; nB := nT
   ELSEIF nI = 3 ; nR := nP ; nG := nQ ; nB := nV
   ELSEIF nI = 4 ; nR := nT ; nG := nP ; nB := nV
   ELSE          ; nR := nV ; nG := nP ; nB := nQ
   ENDIF
ENDIF
RETURN { Round( nR * 255, 0 ), Round( nG * 255, 0 ), Round( nB * 255, 0 ) }

**************************************************
FUNCTION BmpCropResize( oBmp, aRect, aTargetSize )
**************************************************
LOCAL nSizeX   := aRect[ 3 ] - aRect[ 1 ]
LOCAL nSizeY   := aRect[ 4 ] - aRect[ 2 ]
LOCAL oTargetPS:= XbpPresSpace():new():create() 
LOCAL oBitmap  := XbpBitmap():new():create()

oBitmap:presSpace( oTargetPS ) 
oBitmap:make( aTargetSize[ 1 ], aTargetSize[ 2 ] )
oBmp:draw( oTargetPS, { 0, 0, aTargetSize[ 1 ], aTargetSize[ 2 ] }, aRect,, 4 )
oTargetPS:destroy()
RETURN oBitmap

***********************************************
DLL FDLIB IMPORT fdlib_detectfaces AS VOID    ;
    PARAM imagedata                AS LPCHAR, ;
    PARAM imagewidth               AS INT,    ;
    PARAM imageheight              AS INT,    ;
    PARAM threshold                AS INT     ;
SYMBOL "fdlib_detectfaces"
***********************************************

**********************************************
DLL FDLIB IMPORT fdlib_getndetections AS INT ;
SYMBOL "fdlib_getndetections"
**********************************************

*************************************************
DLL FDLIB IMPORT fdlib_getdetection AS VOID     ;
    PARAM nr                        AS INT,     ;
    PARAM xpos                      AS LPINT32, ;
    PARAM ypos                      AS LPINT32, ;
    PARAM width                     AS LPINT32  ;
SYMBOL "fdlib_getdetection"
*************************************************

*********************************************************
STATIC PROCEDURE ShowBmpGrayStep( cGrayData, oBmp, oXbp )  // For test purpose, conditionally used if SHOW_GRAY_STEPS is defined
*********************************************************
LOCAL nCont, oNewBmp
LOCAL nLen   := oBmp:xSize * oBmp:ysize
LOCAL cBuffer:= Left( oBmp:setBuffer(), oBmp:bufferOffset )

FOR nCont:= 1 TO nLen
   cBuffer += Replicate( cGrayData[ nCont ], 3 ) + Chr( 0 )
NEXT
oNewBmp:= XbpBitmap():new():create()
   oNewBmp:setBuffer( cBuffer )

oXbp:setCaption( oNewBmp )
RETURN

**************************************
STATIC FUNCTION BmpCrop( oBmp, aRect )  // For test purpose, not used in this sample
**************************************
LOCAL cRowPadding, nRowPixels, nXPixelsCrop, nXOffset, nRow, oNewBmp
LOCAL nXSize    := aRect[ 3 ] - aRect[ 1 ]
LOCAL nYSize    := aRect[ 4 ] - aRect[ 2 ]
LOCAL cBuffer   := SubStr( oBmp:setBuffer(), oBmp:bufferOffset + 1 )
LOCAL cNewBuffer:= Left( oBmp:setBuffer(), oBmp:bufferOffset )
LOCAL nBits     := oBmp:bits

cNewBuffer:= Left( cNewBuffer, 4 ) + U2Bin( nXSize ) + U2Bin( nYSize ) + SubStr( cNewBuffer, 13 )

IF nBits >= 24
   nBits/= 8
ELSEIF nBits >= 1 .AND. nBits < 24
   nBits:= 1
ENDIF

cRowPadding := Replicate( Chr( 0 ), nXSize * nBits % 4 )
nRowPixels  := oBmp:xSize * nBits + oBmp:xSize * nBits % 4
nXPixelsCrop:= nXSize * nBits
nXOffset    := aRect[ 1 ] * nBits

FOR nRow:= aRect[ 2 ] TO aRect[ 4 ]
   cNewBuffer += SubStr( cBuffer, 1 + nRowPixels * ( nRow - 1 ) + nXOffset, nXPixelsCrop ) + cRowPadding
NEXT
oNewBmp:= XbpBitmap():new():create()
   oNewBmp:setBuffer( cNewBuffer )

RETURN oNewBmp

********************************************
STATIC FUNCTION OtsuEqual( aGrayData, aEQU )  // For test purpose, not used in this sample
********************************************
LOCAL aHistogram := AFill( Array( 256 ), 0 )
LOCAL nLen       := Len( aGrayData )
LOCAL cGrayData  := Space( nLen )
LOCAL nSum       := 0
LOCAL nSumB      := 0
LOCAL nWB        := 0
LOCAL nWF        := 0
LOCAL nMax       := 0
LOCAL nBetween   := 0
LOCAL nThreshold1:= 0
LOCAL nThreshold2:= 0
LOCAL nCont, nAsc, nMB, nMF, nThreshold

FOR nCont:= 1 TO nLen
   nAsc:= aEQU[ aGrayData[ ncont ] + 1 ]
   aGrayData[ nCont ]:= nAsc
   aHistogram[ nAsc + 1 ]++
NEXT
FOR nCont:= 1 TO 256 ; nSum += nCont * aHistogram[ nCont ] ; NEXT

FOR nCont:= 1 TO 256
   nWB += aHistogram[ nCont ]
   IF nWB > 0
      nWF:= nLen - nWB
      IF nWF = 0
         EXIT
      ENDIF
      nSumB   += nCont * aHistogram[ nCont ]
      nMB     := nSumB / nWB
      nMF     := ( nSum - nSumB ) / nWF
      nBetween:= nWB * nWF * ( nMB - nMF ) ** 2

      IF nBetween = nMax
         nThreshold1:= nCont
      ELSEIF nBetween > nMax
         nThreshold1:= nCont
         nThreshold2:= nCont
         nMax       := nBetween           
      ENDIF
    ENDIF
NEXT
nThreshold:= Int( ( nThreshold1 + nThreshold2 ) / 2 )

FOR nCont:= 1 TO nLen ; cGrayData[ nCont ]:= IIF( aGrayData[ nCont ] < nThreshold, Chr( 0 ), Chr( 255 ) ) ; NEXT
RETURN cGrayData
